from pathlib import Path

# we should make this a proper package that loads examples
# similar how they handle it in xarray:
# https://github.com/pydata/xarray/blob/main/xarray/tutorial.py
# Currently there are simple paths to the examples which are then easily loaded in the notebooks

_EXAMPLES = Path(__file__).parent
wind_energy_ontology = _EXAMPLES / "wind-energy.owl"
